﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=9.8
@EndOfDesignText@
'SimpleMixer Requires all Files to be formatted 16bit sample size, Stereo
'44100 sample rate is recommended as most soundcards support that, other sample rates may cause problems on different systems
'which this demo should catch and provide an error message.  For your app, you would need to work out how to treat these issues.
'SimpleMixer provide callbacks for WarningClipValue, Position Events and Sound Events which can be added with the appropriate Add method on SimpleMixer
'PosEvent is called after processing all tracks and provides the current PlayPosition of the track.
'ClipEvent is called when the level of the track, and/or all tracks is above the Threshold requestd in setWarningClipValue.  29000 by default.
'See Full Mixer demo for more information and use.
'SoundEvent is called whrn a sound above the Sound Threshold is found. Default is 500.
'To access these events you need the appropriate Add event on the SimpleMixer.
'If the SoundEvent Wav is null, then it is the Master(Output) value that has risen above the threshold

#Event: ClipEvent(Wav as WavRandomAccessFile,Value As Short)
#Event: PosEvent(PlayPos As Long)
#Event: SoundEvent(Wav as WavRandomAccessFile,Maxvalue As Short)
Sub Class_Globals
	Private fx As JFX
	Private mFiles As List
	Private OutDevice As JavaObject
	Private SDLW As SourceDataLineWrapper
	Private MaxBufferSize As Int = 4096
	Private BufferSize As Int
	Private StopPlaying As Boolean
	Private mPause As Boolean
	Private ReportPos As Boolean
	Private PosCallback As Object
	Private PosEventName As String
	Private ReportClip As Boolean
	Private ClipCallback As Object
	Private ClipEventName As String
	Private ReportSound As Boolean
	Private SoundCallback As Object
	Private SoundEventName As String
	Private SoundThreshold As Short = 500
	Private PlayPos As Long
	Private mIsPlaying As Boolean
	Private mMasterGain As Float = 1
	Private WarningClipValue As Short = 29000 'arbitrary value after which clipping is likely to occur adding digital artefacts to the sound
	Private FirstTime As Boolean
	
	Private MixAudioFormat As JavaObject
	Private MixSampleRateInHz As Float
	Private MixSampleSizeInBits As Int
	Private MixChannelConfig As Int
End Sub


'Initializes the object. You can add parameters to this method if needed.
Public Sub Initialize
	mFiles.Initialize
End Sub

Public Sub AddPositionListener(Callback As Object,EventName As String)
	PosCallback = Callback
	PosEventName = EventName
	If SubExists(PosCallback,PosEventName & "_PosEvent") Then ReportPos = True
End Sub

Public Sub AddClipListener(Callback As Object,EventName As String)
	ClipCallback = Callback
	ClipEventName = EventName
	If SubExists(ClipCallback,ClipEventName & "_ClipEvent") Then ReportClip = True
End Sub

Public Sub AddSoundListener(Callback As Object,EventName As String)
	SoundCallback = Callback
	SoundEventName = EventName
	If SubExists(SoundCallback,SoundEventName & "_SoundEvent") Then ReportSound = True
End Sub

Public Sub AddFile(Folder As String, fileName As String) As Boolean
	Dim Wav As WavRandomAccessFile
	Wav.Initialize(Folder,fileName)
	Return AddFile2(Wav)
End Sub


Public Sub AddFile2(Wav As WavRandomAccessFile) As Boolean
	
	If Wav.SamplesizeInBits <> 16 Or Wav.ChannelConfig <> 2 Then
		Log("This sequencer Cannot play this file. Must be 16bit stereo")
		Return False
	End If
	
	If mFiles.Size = 0 Then
		MixAudioFormat = Wav.AudioFormat
		MixSampleRateInHz = Wav.GetSampleRateInHz
		MixSampleSizeInBits = Wav.SamplesizeInBits
		MixChannelConfig = Wav.ChannelConfig
	Else
		If Wav.AudioFormat.RunMethod("matches",Array(MixAudioFormat)) = False Then
			Log($"File ${Wav.FileName} NOT added - audioFormat not the same as previous added files"$)
			Return False
		End If
	End If
	
	mFiles.Add(Wav)
	
	If OutDevice.IsInitialized = False Then
		Try
			OutDevice = jAudioTrack2_Utils.GetSourceDataLine(Wav.AudioFormat)
			SDLW.Initialize(OutDevice)
		Catch
			Log("Suitable output device not found")
			Log(LastException)
			Return False
		End Try
	End If
	
	Return True
End Sub

Public Sub RemoveFile(Wav As WavRandomAccessFile)
	mFiles.RemoveAt(mFiles.IndexOf(Wav))
	Wav.Close
End Sub

Public Sub Close
	Stop
	
	For Each Wav As WavRandomAccessFile In mFiles
		Wav.close
	Next
End Sub

Public Sub RemoveAllTracks
	For Each Wav As WavRandomAccessFile In mFiles
		Wav.close
	Next
	mFiles.Clear
End Sub

Public Sub getTracks As List
	Return mFiles
End Sub

Public Sub Stop
	StopPlaying = True
End Sub

Public Sub Pause
	mPause = True
End Sub

Public Sub Play
	If mIsPlaying Then Return
	StopPlaying = False
	mPause = False
	PlayDo
End Sub

Public Sub getIsPlaying As Boolean
	Return mIsPlaying
End Sub

Private Sub PlayDo 		'Ignore
	Log("Playing")
	mIsPlaying = True
	
	'check there are tracks to play
	If mFiles.Size = 0 Then
		mIsPlaying = False
		Return
	End If
	
	'Required library
	Dim BC As ByteConverter
	BC.LittleEndian = True
	
	'Working buffer size
	BufferSize = Min(MaxBufferSize,SDLW.getBufferSize)
	
	'Start the source data line
	SDLW.Start

	'Working buffer
	Dim Dat(BufferSize) As Byte
	Dim ThisBufferSize As Int

	'Sample values
	Dim NewSample, Mixed As Short

	'Running total for each buffer operation (all tracks)
	Dim TotalRead As Int
	
	'Position indexes
	Dim i,MixBufferi As Int
	Dim Len As Long
	Dim MaxValue As Short
	
	Dim SoundMap As Map
	SoundMap.Initialize
	
	FirstTime = True
	
	Do While True
		'Maximum length of the last buffer read from each track
		Len = 0
		TotalRead = 0
		
		'Define / Clear the mix buffer for the nexr iteration
		Dim MixBuffer(BufferSize / 2) As Short
		
		
		'Mixing Process
		'For each track:
		'Read each loaded file in sequence as NewSample(2 bytes per sample Left channel then Right channel),
		'Aquire each 16bit sample and convert to float
		'Check the level of each track is not too high (WarningClipValue is an arbitrary value of 29000 & -29000)
		'Add The NewSample to the corresponding value in the MixBuffer(all tracks already processed) into the variable Mixed
		'Check the level of the mixed value is not too high (WarningClipValue is an arbitrary value of 29000 & -29000)
		'Add the Mixed value to the corresponding position of the shortbuffer MixBuffer.
		
		'Once all tracks are added to the MixBuffer:
		'Convert the Short Mixbuffer to bytes and write it to the SourceDataLine
		'Check if the Stop or pause flags are set and exit if needed
		'If the total bytes read from each track is 0, then playing of all tracks is complete so exit.
		'Tidy up setting appropriate flags, stopping the SourceDataLine and return from the sub
		
		For Each Wav As WavRandomAccessFile In mFiles
			MaxValue = 0
			'Avoid reading past the end of the Data Chunk in case the Data Chunk is not the last chunk in the file.
			ThisBufferSize = Max(0, Min(BufferSize , Wav.DataChunkEnd - Wav.PlayPos))
			
			
			TotalRead = TotalRead + Wav.RandomAccessFile.ReadBytes(Dat,0,ThisBufferSize, Wav.DataChunkStart + Wav.PlayPos)
			Len = Max(Len,Dat.Length)
			
			'Play position for each track
			Wav.PlayPos = Wav.PlayPos + Dat.Length
			'Check if track is muted
			For i = 0 To Dat.Length - 1 Step 2
				MixBufferi = i / 2
				'Convert 16 bits to short so they can be added together
				NewSample = BC.ShortsFromBytes(Array As Byte(Dat(i),Dat(i+1)))(0)
				
				'Check input. If the level is too high it could cause clipping.  Just a warning
				If NewSample > WarningClipValue Then
					If ReportClip Then CallSubDelayed3(ClipCallback,ClipEventName & "_ClipEvent",Wav,NewSample)
				End If
				If NewSample < -WarningClipValue Then
					If ReportClip Then CallSubDelayed3(ClipCallback,ClipEventName & "_ClipEvent",Wav,NewSample)
				End If
					
				'Add the NewSample to the MixBuffer corresponding value
				Mixed = (MixBuffer(MixBufferi) + NewSample)

				'Check output level. if the level is too high it could cause clipping.  Just a warning
				If Mixed > WarningClipValue Then
					If ReportClip Then CallSubDelayed3(ClipCallback,ClipEventName & "_ClipEvent",Null,Mixed)
				End If
				If Mixed < -WarningClipValue Then
					If ReportClip Then CallSubDelayed3(ClipCallback,ClipEventName & "_ClipEvent",Null,Mixed)
				End If
					
				'Update the current Mixbuffer
				MixBuffer(MixBufferi) = Mixed
				MaxValue = Max(MaxValue,NewSample)
			Next
			If ReportSound And MaxValue > SoundThreshold Then SoundMap.Put(Wav,MaxValue)
						
			'Free up the gui
			Sleep(0)
			
			'Exit if flag is set
			If StopPlaying Then Exit
		Next

		'Report current position if requested
		If ReportPos Then CallSubDelayed2(PosCallback,PosEventName & "_PosEvent",PlayPos)
		
		'Update play position for reporting (Maximum value)
		PlayPos = PlayPos + Len
		
		'Convert the mixed MixBuffer to bytes and write it to the SourceDataLine
		SDLW.Write(BC.ShortsToBytes(MixBuffer),0,MixBuffer.Length * 2)
		
		If SoundMap.Size > 0 Then
			'Clone the soundmap so it can be cleared for the next iteration.
			Dim SoundMap1 As Map
			SoundMap1.Initialize
			SoundMap1.As(JavaObject).RunMethod("putAll",Array(SoundMap))
			SoundMap.Clear
			CallSubDelayed3(Me,"Report_Sound",IIf(FirstTime,True,False),SoundMap1)
		End If
		FirstTime = False
		
		'Exit if flag is set
		If StopPlaying Or mPause Then Exit
		
		'Will occur when all RandomAccess files are exhausted.
		If TotalRead = 0 Then Exit
	Loop
	
	Log("Stopped")
	
	If StopPlaying Or TotalRead = 0 Then
		'Reset play position to beginning
		setPlayPos(0)
		
		'Report current position if requested.  In this app it resets the position slider play position
		If ReportPos Then CallSubDelayed2(PosCallback,PosEventName & "_PosEvent",PlayPos)
	End If
	
	'Stop and clear the Source Data Line
	SDLW.Stop
	SDLW.Drain
	SDLW.flush
	
	'Reset the playing flag
	mIsPlaying = False
End Sub

Private Sub Report_Sound(tFirstTime As Boolean,SoundMap1 As Map)
	If Not(tFirstTime) Then Sleep(300)
	For Each Wav As WavRandomAccessFile In SoundMap1.Keys
		Dim MaxVal As Short = SoundMap1.Get(Wav)
		CallSubDelayed3(SoundCallback,SoundEventName & "_SoundEvent",Wav,MaxVal)
	Next
End Sub

Public Sub setSoundThreshold(Threshold As Short)
	SoundThreshold = Threshold
End Sub

Public Sub getSoundThreshold As Short
	Return SoundThreshold
End Sub

Public Sub setMaxBufferSize(tBufferSize As Int)
	MaxBufferSize = tBufferSize
End Sub

Public Sub getMaxBufferSize As Int
	Return MaxBufferSize
End Sub

Public Sub setWarningClipValue(Value As Short)
	WarningClipValue = Value
End Sub

Public Sub getWarningClipValue As Short
	Return WarningClipValue
End Sub

Public Sub getmIsPlaying As Boolean
	Return mIsPlaying
End Sub

Public Sub setMasterGain(Gain As Float)
	mMasterGain = Gain
End Sub
'Get/Set Master gain
Public Sub getMasterGain As Float
	Return mMasterGain
End Sub

Public Sub getFiles As List
	Return mFiles
End Sub

'Get the sourceDataLine
Public Sub getSourceDataLine As SourceDataLineWrapper
	Return SDLW
End Sub

'Get/Set the play position
Public Sub setPlayPos(Pos As Long)
	PlayPos = Pos
	For Each Wav As WavRandomAccessFile In mFiles
		Wav.PlayPos = Pos
	Next
End Sub

Public Sub getPlayPos As Long
	Return PlayPos
End Sub

'Get the play duration in Milliseconds
Public Sub getMilliSecondDuration As Long
	Dim Dur As Long
	For Each Wav As WavRandomAccessFile In mFiles
		Dur = Max(Dur,Wav.Duration)
	Next
	Return Dur * 1000
End Sub

'Get/Set the millisecond play position
Public Sub getMillisecondPosition As Long
	Return PlayPos / (MixSampleRateInHz * MixChannelConfig * (MixSampleSizeInBits / 8)) * 1000
End Sub

Public Sub setMillisecondPosition(mSeconds As Long)
	Dim NewPos As Long = Floor((mSeconds / 1000) * MixSampleRateInHz * MixChannelConfig) * (MixSampleSizeInBits / 8)
	PlayPos = NewPos
	
	'Set on each track
	For Each Wav As WavRandomAccessFile In mFiles
		Wav.PlayPos = PlayPos
	Next
End Sub